#pragma once

#include <vector>

#include <RobotAPI/interface/armem/query.h>

#include "../core/Time.h"
#include "../core/ice_conversions.h"

#include "query/detail.h"


namespace armarx::armem::client
{

    class SnapshotsSelector :
        public detail::ChildSelectorOps<SnapshotsSelector, query::EntityQueryPtr>
    {
    public:


        std::vector<query::EntityQueryPtr> build() const
        {
            return _queries;
        }


        SnapshotsSelector& all(bool withData = true)
        {
            auto q = _addQuery<query::entity::All>(withData);
            q.withData = withData;
            return *this;
        }
        SnapshotsSelector& latest(bool withData = true);

        SnapshotsSelector& atTime(Time timestamp, bool withData = true)
        {
            auto q = _addQuery<query::entity::Single>(withData);
            toIce(q.timestamp, timestamp);
            return *this;
        }

        SnapshotsSelector& timeRange(Time min, Time max, bool withData = true)
        {
            auto q = _addQuery<query::entity::TimeRange>(withData);
            toIce(q.minTimestamp, min);
            toIce(q.maxTimestamp, max);
            return *this;
        }

        SnapshotsSelector& indexRange(long first, long last, bool withData = true)
        {
            auto q = _addQuery<query::entity::IndexRange>(withData);
            q.first = first;
            q.last = last;
            return *this;
        }


    protected:

        template <class DerivedQueryT>
        DerivedQueryT& _addQuery(bool withData)
        {
            DerivedQueryT& query = ChildSelectorOps::_addQuery<DerivedQueryT>();
            query.withData = withData;
            return query;
        }

    };


    class EntitiesSelector :
        public detail::InnerSelectorOps<EntitiesSelector, query::ProviderSegmentQueryPtr, SnapshotsSelector>
    {
    public:

        SnapshotsSelector& snapshots()
        {
            return _addChild();
        }


        EntitiesSelector& all()
        {
            auto q = _addQuery<query::provider::All>();
            return *this;
        }

        EntitiesSelector& withName(const std::string& name);
        EntitiesSelector& withNames(const std::vector<std::string>& names);

        EntitiesSelector& withNameMatching(const std::string& regex);


    protected:

        void _setChildQueries(QueryT& query, const std::vector<ChildT::QueryT>& childQueries) const override;
    };

    void EntitiesSelector::_setChildQueries(ChildSelectorOps::QueryT& query, const std::vector<ChildT::QueryT>& childQueries) const
    {
        query->entityQueries = childQueries;
    }



    class ProviderSegmentSelector :
        public detail::InnerSelectorOps<ProviderSegmentSelector, query::CoreSegmentQueryPtr, EntitiesSelector>
    {
    public:

        EntitiesSelector& entities()
        {
            return _addChild();
        }

        ProviderSegmentSelector& all()
        {
            auto q = _addQuery<query::core::All>();
            return *this;
        }

        ProviderSegmentSelector& withName(const std::string& name);
        ProviderSegmentSelector& withNames(const std::vector<std::string>& names);

        ProviderSegmentSelector& withNameMatching(const std::string& regex);

    protected:

        void _setChildQueries(QueryT& query, const std::vector<ChildT::QueryT>& childQueries) const override;
    };

    void ProviderSegmentSelector::_setChildQueries(ChildSelectorOps::QueryT& query, const std::vector<ChildT::QueryT>& childQueries) const
    {
        query->providerSegmentQueries = childQueries;
    }



    class CoreSegmentSelector :
        public detail::InnerSelectorOps<CoreSegmentSelector, query::MemoryQueryPtr, ProviderSegmentSelector>
    {
    public:

        ProviderSegmentSelector& providerSegments()
        {
            return _addChild();
        }

        CoreSegmentSelector& all()
        {
            auto q = _addQuery<query::memory::All>();
            return *this;
        }

        CoreSegmentSelector& withName(const std::string& name);
        CoreSegmentSelector& withNames(const std::vector<std::string>& names);

        CoreSegmentSelector& withNameMatching(const std::string& regex);


    protected:

        void _setChildQueries(QueryT& query, const std::vector<ChildT::QueryT>& childQueries) const override;
    };

    void CoreSegmentSelector::_setChildQueries(query::MemoryQueryPtr& query, const std::vector<ChildT::QueryT>& childQueries) const
    {
        query->coreSegmentQueries = childQueries;
    }


    // ToDo: Make a memory selector (but this level is not in ice, yet)
    class QueryBuilder :
        public detail::ParentSelectorOps<QueryBuilder, CoreSegmentSelector>
    {
    public:

        std::vector<query::MemoryQueryPtr> finalize() const
        {
            std::vector<query::MemoryQueryPtr> childQueries;
            for (const CoreSegmentSelector& child : _children)
            {
                for (const query::MemoryQueryPtr& query : child.build())
                {
                    childQueries.push_back(query);
                }
            }
            return childQueries;
        }


        CoreSegmentSelector& coreSegments()
        {
            return _addChild();
        }

    };

}
