#include "MemoryID.h"

#include <SimoxUtility/algorithm/string.h>

#include "../error/ArMemError.h"


namespace armarx::armem
{

    MemoryID::MemoryID()
    {
    }

    MemoryID::MemoryID(const std::string& string)
    {
        std::vector<std::string> items = simox::alg::split(string, "/", false, false);

        auto it = items.begin();
        if (it == items.end())
        {
            return;
        }
        this->memoryName = *it;

        if (++it == items.end())
        {
            return;
        }
        this->coreSegmentName = *it;

        if (++it == items.end())
        {
            return;
        }
        this->providerSegmentName = *it;

        if (++it == items.end())
        {
            return;
        }
        this->entityName = *it;

        if (++it == items.end())
        {
            return;
        }
        this->timestamp = timestampFromStr(*it);

        if (++it == items.end())
        {
            return;
        }
        this->instanceIndex = instanceIndexFromStr(*it);
    }

    std::string MemoryID::str() const
    {
        std::vector<std::string> items = getAllItems();
        while (items.size() > 0 && items.back().empty())
        {
            items.pop_back();
        }

        return simox::alg::join(items, "/", false, false);
    }


    MemoryID MemoryID::fromString(const std::string& string)
    {
        return MemoryID(string);
    }

    std::vector<std::string> MemoryID::getItems() const
    {
        std::vector<std::string> items;

        items.push_back(memoryName);

        if (!hasCoreSegmentName())
        {
            return items;
        }
        items.push_back(coreSegmentName);

        if (!hasProviderSegmentName())
        {
            return items;
        }
        items.push_back(providerSegmentName);

        if (!hasEntityName())
        {
            return items;
        }
        items.push_back(entityName);

        if (!hasTimestamp())
        {
            return items;
        }
        items.push_back(timestampStr());

        if (!hasInstanceIndex())
        {
            return items;
        }
        items.push_back(instanceIndexStr());

        return items;
    }

    std::vector<std::string> MemoryID::getAllItems() const
    {
        return
        {
            memoryName, coreSegmentName, providerSegmentName, entityName,
            timestampStr(), instanceIndexStr()
        };
    }

    MemoryID MemoryID::getMemoryID() const
    {
        MemoryID id;
        id.memoryName = memoryName;
        return id;
    }

    MemoryID MemoryID::getCoreSegmentID() const
    {
        MemoryID id = getMemoryID();
        id.providerSegmentName = providerSegmentName;
        return id;
    }

    MemoryID MemoryID::getProviderSegmentID() const
    {
        MemoryID id = getCoreSegmentID();
        id.coreSegmentName = coreSegmentName;
        return id;
    }

    MemoryID MemoryID::getEntityID() const
    {
        MemoryID id = getProviderSegmentID();
        id.entityName = entityName;
        return id;
    }

    MemoryID MemoryID::getEntitySnapshotID() const
    {
        MemoryID id = getEntityID();
        id.timestamp = timestamp;
        return id;
    }

    MemoryID MemoryID::getEntityInstanceID() const
    {
        MemoryID id = getEntitySnapshotID();
        id.instanceIndex = instanceIndex;
        return id;
    }

    void MemoryID::setMemoryID(const MemoryID& id)
    {
        memoryName = id.memoryName;
    }

    void MemoryID::setCoreSegmentID(const MemoryID& id)
    {
        setMemoryID(id);
        coreSegmentName = id.coreSegmentName;
    }

    void MemoryID::setProviderSegmentID(const MemoryID& id)
    {
        setCoreSegmentID(id);
        providerSegmentName = id.providerSegmentName;
    }

    void MemoryID::setEntityID(const MemoryID& id)
    {
        setProviderSegmentID(id);
        entityName = id.entityName;
    }

    void MemoryID::setEntitySnapshotID(const MemoryID& id)
    {
        setEntityID(id);
        timestamp = id.timestamp;
    }

    void MemoryID::setEntityInstanceID(const MemoryID& id)
    {
        setEntitySnapshotID(id);
        instanceIndex = id.instanceIndex;
    }

    MemoryID MemoryID::withMemoryName(const std::string& name) const
    {
        MemoryID id = *this;
        id.memoryName = name;
        return id;
    }

    MemoryID MemoryID::withCoreSegmentName(const std::string& name) const
    {
        MemoryID id = *this;
        id.coreSegmentName = name;
        return id;
    }

    MemoryID MemoryID::withProviderSegmentName(const std::string& name) const
    {
        MemoryID id = *this;
        id.providerSegmentName = name;
        return id;
    }

    MemoryID MemoryID::withEntityName(const std::string& name) const
    {
        MemoryID id = *this;
        id.entityName = name;
        return id;
    }

    MemoryID MemoryID::withTimestamp(Time time) const
    {
        MemoryID id = *this;
        id.timestamp = time;
        return id;
    }

    MemoryID MemoryID::withInstanceIndex(int index) const
    {
        MemoryID id = *this;
        id.instanceIndex = index;
        return id;
    }

    std::string MemoryID::timestampStr() const
    {
        return hasTimestamp() ? std::to_string(timestamp.toMicroSeconds()) : "";
    }

    std::string MemoryID::instanceIndexStr() const
    {
        return hasInstanceIndex() ? std::to_string(instanceIndex) : "";
    }

    IceUtil::Time MemoryID::timestampFromStr(const std::string& string)
    {
        return Time::microSeconds(parseInteger(string, "timestamp"));
    }

    int MemoryID::instanceIndexFromStr(const std::string& string)
    {
        return int(parseInteger(string, "instance index"));
    }

    long long MemoryID::parseInteger(const std::string& string, const std::string& semanticName)
    {
        try
        {
            return std::stoll(string);
        }
        catch (const std::invalid_argument&)
        {
            throw error::ParseIntegerError(string, semanticName);
        }
        catch (const std::out_of_range&)
        {
            throw error::ParseIntegerError(string, semanticName);
        }
    }

    std::ostream& operator<<(std::ostream& os, const MemoryID id)
    {
        return os << "'" << id.str() << "'";
    }



}
