#pragma once

#include <memory>
#include <vector>

#include "../core/MemoryID.h"
#include "../core/Time.h"

#include "EntityInstance.h"
#include "detail/MemoryItem.h"
#include "detail/MemoryContainer.h"


namespace armarx::armem
{

    class EntitySnapshot;
    using EntitySnapshotPtr = std::unique_ptr<EntitySnapshot>;


    /**
     * @brief Data of an entity at one point in time.
     */
    class EntitySnapshot :
        public detail::MemoryContainer<std::vector<EntityInstancePtr>, EntitySnapshot>
    {
        using Base = detail::MemoryContainer<std::vector<EntityInstancePtr>, EntitySnapshot>;

    public:

        EntitySnapshot();
        EntitySnapshot(Time time, const MemoryID& parentID = {});
        EntitySnapshot(const MemoryID& id);

        /// Copy the instances from `other` to this.
        EntitySnapshot(const EntitySnapshot& other);
        /// Copy the instances from `other` to this.
        EntitySnapshot& operator=(const EntitySnapshot& other);


        inline Time& time()
        {
            return id().timestamp;
        }
        inline const Time& time() const
        {
            return id().timestamp;
        }


        void update(const InternalEntityUpdate& update,
                    std::optional<MemoryID> parentID = std::nullopt);


        /**
         * @brief Get the given instance.
         * @param index The instance's index.
         * @return The instance.
         * @throw `armem::error::MissingEntry` If the given index is invalid.
         */
        EntityInstance& getInstance(int index);
        const EntityInstance& getInstance(int index) const;
        /**
         * @brief Get the given instance.
         * @param index The instance's index.
         * @return The instance.
         * @throw `armem::error::MissingEntry` If the given index is invalid.
         * @throw `armem::error::InvalidMemoryID` If memory ID does not have an instance index.
         */
        EntityInstance& getInstance(const MemoryID& id);
        const EntityInstance& getInstance(const MemoryID& id) const;


        // MemoryItem interface
    public:
        std::string getKeyString() const override;
        std::string getLevelName() const override;


        // MemoryContainer interface
    protected:
        ContainerT& _underlyingContainer() override;
        void _copyElements(EntitySnapshot& other, bool withData) const override;


    public:

        std::vector<EntityInstancePtr> instances;

    };

}
