/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// Header
#include "AronDataNavigatorReaderToken.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>

namespace armarx
{
    namespace aron
    {
        namespace io
        {
            // constructors
            AronDataNavigatorReaderToken::AronDataNavigatorReaderToken(const datanavigator::AronDataNavigatorPtr& data) :
                data(data),
                currentIndex(0),
                maxIndex(0),
                currentKey("")
            {
                auto desc = getDescriptor();
                switch (desc)
                {
                case eAronDict:
                {
                    allKeys = datanavigator::AronDictDataNavigator::DynamicCast(data)->getAllKeys();
                    break;
                }
                case eAronList:
                {
                    maxIndex = datanavigator::AronListDataNavigator::DynamicCast(data)->childrenSize();
                    break;
                }
                default:
                    throw exception::AronDataDescriptorNotValidException("AronDataNavigatorReaderToken", "AronDataNavigatorReaderToken", "Received an invalid instance for a reader token.", desc);
                }
            }

            // public member functions
            AronDataDescriptor AronDataNavigatorReaderToken::getDescriptor() const
            {
                return data->getDescriptor();
            }

            bool AronDataNavigatorReaderToken::checkedAllChildren() const
            {
                return currentIndex >= data->childrenSize();
            }

            unsigned int AronDataNavigatorReaderToken::getCurrentIndex() const
            {
                return currentIndex;
            }

            // Dict member functions
            std::string AronDataNavigatorReaderToken::getCurrentKey() const
            {
                auto desc = getDescriptor();
                if (desc != AronDataDescriptor::eAronDict)
                {
                    throw exception::AronDataDescriptorNotValidException("AronDataNavigatorReaderToken", "getCurrentKey", "Cant get key of a non-dict token", desc);
                }

                // ugly workaround
                if(currentKey == "") // probably in a dict where the caller doues not know which keys are there
                {
                    return allKeys[getCurrentIndex()];
                }
                return currentKey;
            }

            // Dict member functions
            void AronDataNavigatorReaderToken::setCurrentKey(const std::string& k)
            {
                if(k.empty())
                {
                    throw exception::AronException("AronDataNavigatorReaderToken", "setCurrentKey", "Trying to set an empty key!");
                }
                currentKey = k;
            }

            // virtual definitions
            datanavigator::AronDataNavigatorPtr AronDataNavigatorReaderToken::getElementAndIncreaseCnt()
            {
                auto desc = getDescriptor();
                switch (desc)
                {
                    case eAronDict:
                    {
                        datanavigator::AronDictDataNavigatorPtr casted = datanavigator::AronDictDataNavigator::DynamicCast(data);
                        datanavigator::AronDataNavigatorPtr ret = casted->getElement(getCurrentKey());
                        currentIndex++;
                        return ret;
                    }
                    case eAronList:
                    {
                        datanavigator::AronListDataNavigatorPtr casted = datanavigator::AronListDataNavigator::DynamicCast(data);
                        datanavigator::AronDataNavigatorPtr ret = casted->getElement(getCurrentIndex());
                        currentIndex++;
                        return ret;
                    }
                    default:
                        throw exception::AronDataDescriptorNotValidException("AronDataNavigatorReaderToken", "getElementAndIncreaseCnt", "Could not resove a type of a navigator. Allowed are only container and complex types due to performance", desc);
                }
            }
        }
    }
}
